@@ -18,7 +18,6 @@
 
 #include "JCCEnv.h"
 
-
 #if defined(_MSC_VER) || defined(__WIN32)
 _DLL_EXPORT DWORD VM_ENV = 0;
 #else
@@ -111,20 +110,25 @@ void JCCEnv::set_vm(JavaVM *vm, JNIEnv *vm_env)
         vm_env->GetMethodID(_obj, "getClass",
                             "()Ljava/lang/Class;");
 
-    if (vm_env->GetVersion() >= 0x00010005)
+
+    jclass iterable = vm_env->FindClass("java/lang/Iterable");
+
+    if (iterable == NULL) /* JDK < 1.5 */
+    {
+        vm_env->ExceptionClear();
+        _mids[mid_iterator] = NULL;
+        _mids[mid_iterator_next] = NULL;
+    }
+    else
     {
         _mids[mid_iterator] =
-            vm_env->GetMethodID(vm_env->FindClass("java/lang/Iterable"),
+            vm_env->GetMethodID(iterable,
                                 "iterator", "()Ljava/util/Iterator;");
         _mids[mid_iterator_next] =
             vm_env->GetMethodID(vm_env->FindClass("java/util/Iterator"),
                                 "next", "()Ljava/lang/Object;");
     }
-    else
-    {
-        _mids[mid_iterator] = NULL;
-        _mids[mid_iterator_next] = NULL;
-    }
+
 
     _mids[mid_enumeration_nextElement] =
         vm_env->GetMethodID(vm_env->FindClass("java/util/Enumeration"),
@@ -156,6 +160,24 @@ void JCCEnv::set_vm(JavaVM *vm, JNIEnv *vm_env)
                             "shortValue", "()S");
 }
 
+int JCCEnv::attachCurrentThread(char *name, int asDaemon)
+{
+    JNIEnv *jenv = NULL;
+    JavaVMAttachArgs attach = {
+        JNI_VERSION_1_4, name, NULL
+    };
+    int result;
+
+    if (asDaemon)
+        result = vm->AttachCurrentThreadAsDaemon((void **) &jenv, &attach);
+    else
+        result = vm->AttachCurrentThread((void **) &jenv, &attach);
+
+    set_vm_env(jenv);
+
+    return result;
+}
+
 #if defined(_MSC_VER) || defined(__WIN32)
 
 void JCCEnv::set_vm_env(JNIEnv *vm_env)
@@ -314,7 +336,20 @@ jobject JCCEnv::deleteGlobalRef(jobject obj, int id)
                 {
                     if (iter->second.count == 1)
                     {
-                        get_vm_env()->DeleteGlobalRef(iter->second.global);
+                        JNIEnv *vm_env = get_vm_env();
+
+                        if (!vm_env)
+                        {
+                            /* Python's cyclic garbage collector may remove
+                             * an object inside a thread that is not attached
+                             * to the JVM. This makes sure the JVM doesn't
+                             * segfault.
+                             */
+                            attachCurrentThread(NULL, 0);
+                            vm_env = get_vm_env();
+                        }
+
+                        vm_env->DeleteGlobalRef(iter->second.global);
                         refs.erase(iter);
                     }
                     else
@@ -710,10 +745,11 @@ void JCCEnv::setClassPath(const char *classPath)
     jmethodID ma = vm_env->GetMethodID(_ucl, "addURL", "(Ljava/net/URL;)V");
 #if defined(_MSC_VER) || defined(__WIN32)
     char *pathsep = ";";
+    char *path = _strdup(classPath);
 #else
     char *pathsep = ":";
-#endif
     char *path = strdup(classPath);
+#endif
 
     for (char *cp = strtok(path, pathsep);
          cp != NULL;
